#include "DCT.h"

int DCT_initialized = 0;
float DCT_bases[BLOCK_SIZEY][BLOCK_SIZEX];
float IDCT_bases[BLOCK_SIZEY][BLOCK_SIZEX];
fRGB DCT_temp[BLOCK_SIZEY][BLOCK_SIZEX];

float DCT_threshold[BLOCK_SIZEY][BLOCK_SIZEX] =
{
	8.0f,   5.5f,   5.0f,   8.0f,    12.0f,   20.0f,   25.5f,   30.5f,
	6.0f,   6.0f,   7.0f,   9.5f,    13.0f,   29.0f,   30.0f,   27.5f,
	7.0f,   6.5f,   8.0f,   12.0f,   20.0f,   28.5f,   34.5f,   28.0f,
	7.0f,   8.5f,   11.0f,  14.5f,   25.5f,   43.5f,   40.0f,   31.0f,
	9.0f,   11.0f,  18.5f,  28.0f,   34.0f,	  54.5f,   51.5f,   38.5f,
	12.0f,  17.5f,  27.5f,  32.0f,   40.5f,   52.0f,   56.5f,   46.0f,
	24.5f,  32.0f,  39.0f,  43.5f,   51.5f,   60.5f,   60.0f,   50.5f,
	36.0f,  46.0f,  47.5f,  49.0f,   56.0f,   50.0f,   51.5f,   59.5f
};


int zigzag[BLOCK_SIZEX * BLOCK_SIZEY] =
{
	0,   1,  5,  6, 14, 15, 27, 28,
	2,   4,  7, 13, 16, 26, 29, 42,
	3,   8, 12, 17, 25, 30, 41, 43,
	9,  11, 18, 24, 31, 40, 44, 53,
	10, 19, 23, 32, 39, 45, 52, 54,
	20, 22, 33, 38, 46, 51, 55, 60,
	21, 34, 37, 47, 50, 56, 59, 61,
	35, 36, 48, 49, 57, 58, 62, 63
	
}; 


void init_DCT()
{
	if (DCT_initialized)
		return;

	unsigned int i, j;

	for (j = 0; j < BLOCK_SIZEX; j++) 
	{
		for (i = 0; i < BLOCK_SIZEX; i++) 
		{
			DCT_bases[i][j] = (float) cos(((2 * i + 1) * j) * 3.14159265358979323 / (2.0 * BLOCK_SIZEX));		// assumes BLOCK_SIZEX == BLOCK_SIZEY, otherwise must use two DCT bases, one for each direction
			IDCT_bases[i][j] = DCT_bases[i][j] * (j == 0 ? 0.7071f : 1.0f);
		}
	}

	DCT_initialized = 1;
}



void DCT(floatimage &img)
{
	unsigned int u, v;
	unsigned int maxu = img.sizex / BLOCK_SIZEX;
	unsigned int maxv = img.sizey / BLOCK_SIZEY;

	init_DCT();

	for (v = 0; v < maxv; v++) 
	{
		for (u = 0; u < maxu; u++) 
		{
			DCTBlock(img, u * BLOCK_SIZEX, v * BLOCK_SIZEY);
		}
	}
}


void DCTBlock(floatimage &img, int u0, int v0)
{
	unsigned int u,v,i,j;
	fRGB sum;

	for (v = 0; v < BLOCK_SIZEY; v++)
	{
		memcpy(DCT_temp[v], &img.line[v0 + v][u0], sizeof(fRGB) * BLOCK_SIZEX);
	}

	float av = 0.7071f;
	for (v = 0; v < BLOCK_SIZEY; v++) 
	{
		for (u = 0; u < BLOCK_SIZEX; u++) 
		{
			sum.Set(0.0f, 0.0f, 0.0f);
			for (j = 0; j < BLOCK_SIZEY; j++) 
			{
				for (i = 0; i < BLOCK_SIZEX; i++) 
				{
					float m = DCT_bases[i][u] * DCT_bases[j][v];
					sum.r += DCT_temp[j][i].r * m;
					sum.g += DCT_temp[j][i].g * m;
					sum.b += DCT_temp[j][i].b * m;
				}
			}
			float auv = 0.25f * av * ((u==0) ? 0.7071f : 1.f);

			img.line[v + v0][u + u0].Set(sum.r * auv, sum.g * auv, sum.b * auv);
		}
		av = 1.0f;
	}
}



void IDCT(floatimage &img)
{
	unsigned int u, v;
	unsigned int maxu = img.sizex / BLOCK_SIZEX;
	unsigned int maxv = img.sizey / BLOCK_SIZEY;

	init_DCT();

	for (v = 0; v < maxv; v++) 
	{
		for (u = 0; u < maxu; u++) 
		{
			IDCTBlock(img, u * BLOCK_SIZEX, v * BLOCK_SIZEY);
		}
	}
}


void IDCTBlock(floatimage &img, int u0, int v0)
{
	unsigned int u,v,i,j;
	fRGB sum;

	for (v = 0; v < BLOCK_SIZEY; v++)
	{
		memcpy(DCT_temp[v], &img.line[v0 + v][u0], sizeof(fRGB) * BLOCK_SIZEX);
	}

	for (v = 0; v < BLOCK_SIZEY; v++) 
	{
		for (u = 0; u < BLOCK_SIZEX; u++) 
		{
			sum.Set(0.0f, 0.0f, 0.0f);
			for (j = 0; j < BLOCK_SIZEY; j++) 
			{
				for (i = 0; i < BLOCK_SIZEX; i++) 
				{
					float m = IDCT_bases[u][i] * IDCT_bases[v][j];
					sum.r += DCT_temp[j][i].r * m;
					sum.g += DCT_temp[j][i].g * m;
					sum.b += DCT_temp[j][i].b * m;
				}
			}
			img.line[v + v0][u + u0].Set(sum.r * 0.25f, sum.g * 0.25f, sum.b * 0.25f);
		}
	}
}


void ZigZag(floatimage &img)
{
	fRGB *temp = new fRGB[img.sizex * img.sizey];
	memcpy(temp, img.line[0], img.sizex * img.sizey * sizeof(fRGB));

	int cur = 0;
	for (int by = 0; by < img.sizey / BLOCK_SIZEY; by++)
	{
		for (int bx = 0; bx < img.sizex / BLOCK_SIZEX; bx++)
		{
			int i = 0;
			for (int y = 0; y < BLOCK_SIZEY; y++)
			{
				for (int x = 0; x < BLOCK_SIZEX; x++)
				{
					int ax = bx * BLOCK_SIZEX + x;
					int ay = by * BLOCK_SIZEY + y;
					img.line[0][cur + zigzag[i]] = temp[ay * img.sizex + ax];
					i++;
				}
			}
			cur += BLOCK_SIZEX * BLOCK_SIZEY;
		}
	}


	// SWAP THE DC COMPONENTS TO THE FIRST 8 BYTES OF THE REFLECTANCE MAP DATA - hack alert
	
	cur = 1;			
	for (int x = BLOCK_SIZEX * BLOCK_SIZEY; x < img.sizex * img.sizey; x += BLOCK_SIZEX * BLOCK_SIZEY)
	  {
	    fRGB temp = img.line[0][cur];
	    img.line[0][cur] = img.line[0][x];
	    img.line[0][x] = temp;
	    cur++;
	  }


	delete temp;
}
