#include "Render.h"


int MonstaRenderer::StartCalculation(floatimage &environment, int cur, int max)
{	
	mutex_lock(&thread_state_mutex);
	if (thread_state)					// some threads still calculating
	{
		mutex_unlock(&thread_state_mutex);
		return 0;
	}
	
	if (cur == 0)		// recalculate environment DCT and RLE
	{
		CalculateEnvironment(environment);
		left = 0;
		right = data->sizex - 1;
	}
	else
	{
		left  = (int) (((float) cur + 0) * (data->sizex) / max);
		right = (int) (((float) cur + 1) * (data->sizex) / max) - 1;
	}

	thread_state = (1 << thread_count) - 1;
	mutex_unlock(&thread_state_mutex);
	cond_broadcast(&start_calculation_signal);

	return 1;
}

THREAD_PROTOTYPE(calc_img_thread,lpParameter)
{
	thread_params *p = (thread_params *) lpParameter;
	MonstaRenderer *m = p->rma;
	
	mutex_t *state_mutex = &m->thread_state_mutex;
	cond_t *wait_cond = &m->start_calculation_signal;

	int my_flag = (1 << p->thread_number);

	// thread_priority_low();

	while (true)
	{
		mutex_lock(state_mutex);
		m->thread_state &= ~my_flag;	// I am waiting! (clear the relevant bit)

		while (!(m->thread_state & my_flag))	// while I am waiting
		{
			cond_wait(wait_cond, state_mutex);	 // Release mutex and wait for signal simultaneously
		}
		mutex_unlock(state_mutex);

		if (m->thread_die_flag)
			break;

		int start  = (int) (((float) p->draw_strip + 0) * (m->data->sizey) / p->max_strips);
		int end    = (int) (((float) p->draw_strip + 1) * (m->data->sizey) / p->max_strips) - 1;

		m->calculateLines(start, end);
	}

	mutex_lock(state_mutex);
	m->thread_state &= ~my_flag;		// signal I am dead
	mutex_unlock(state_mutex);
#ifdef _WIN32
	return 1;
#endif
}


void MonstaRenderer::calculateLines(int start, int end)
{
	for (int y = start; y <= end; y++)
	{
		for (int x = left; x <= right; x++)
		{

		  if(data->redRLE(x, y) == NULL)
		    finished_img->line[y][x].rgbRed =
		      finished_img->line[y][x].rgbGreen =
		      finished_img->line[y][x].rgbBlue = 0;

		  else {

#ifdef IMAGE_RGB
		  finished_img->line[y][x].rgbBlue =
#else
		    finished_img->line[y][x].rgbRed =
#endif
				computePixelChannel((char *) data->redRLE(x, y), env_dct_int[0], data->max_phi * data->max_theta);

			finished_img->line[y][x].rgbGreen =
				computePixelChannel((char *) data->greenRLE(x, y), env_dct_int[1], data->max_phi * data->max_theta);

#ifdef IMAGE_RGB
		  finished_img->line[y][x].rgbRed =
#else
		    finished_img->line[y][x].rgbBlue =
#endif
				computePixelChannel((char *) data->blueRLE(x, y), env_dct_int[2], data->max_phi * data->max_theta);
		  }
		}
	}
}

unsigned char MonstaRenderer::computePixelChannel(char *rle, int *mult, int count)
{
	int sum = 0;
	int *max = &mult[count];
	
	
	char temp = 0;

	int c = count / (BLOCK_SIZEX * BLOCK_SIZEY);		// number of DC components;

	while (c)											// sum DC components
	{
		sum += ((unsigned char) *rle++) * (*mult++);
		c--;
	}

	while (mult < max)									// sum RLE AC components
	{
		int nzcount = (unsigned char) *rle++;

		while (nzcount)
		{
			sum += (*rle++) * (*mult++);
			nzcount--;
		};	
	
		//if (mult >= max)
		//break;
	
		mult += (unsigned char) (*rle++);
	}

	sum >>= exposure;

      	return (sum < 0 ? 0 : sum > 4096+4096+1937 ? 252 : linear_to_gamma_table[sum]);
}


void MonstaRenderer::InitThreads(int num_threads)
{
	// Initialize Threads

	thread_die_flag = 0;

	cond_init(&start_calculation_signal);
	mutex_init(&thread_state_mutex);

	mutex_lock(&thread_state_mutex);
	thread_state = 0;

	float start = 0.0f;
	thread_count = 0;
	int x;

	for (x = 0; x < num_threads; x++)
	{
		thread_state |= (1 << x);		// thread is currently NOT waiting, set the bit

		tparams[x].rma = this;

		tparams[x].draw_strip = x;
		tparams[x].max_strips = num_threads;

//		tparams[x].start_ofs = (int) start;		
//		start += (float) data->sizey / (float) num_threads;
//		tparams[x].end_ofs = ((int) start) - 1;

		tparams[x].thread_number = x;

		thread_start(&thread[x], calc_img_thread, &tparams[x]);

		thread_count++;
	}

//	tparams[0].start_ofs = 0;
//	tparams[x-1].end_ofs = data->sizey - 1;

	mutex_unlock(&thread_state_mutex);

#ifdef _WIN32 // let the other threads schedule
	Sleep(0);
#else
	sleep(0);
#endif
}




void MonstaRenderer::CalculateEnvironment(floatimage &environment)
{
  int y;

	// calculate env_dct
	env_dct.copy(environment);
	
	// scale based on phi
	for (y = 0; y < env_dct.sizey; y++)	{
		float phiscale = sin((y + 0.5) * 3.14159265358979323 / env_dct.sizey);

		for (int x = 0; x < env_dct.sizex; x++)
		{
			env_dct.line[y][x] *= phiscale;
		}
	}

	::DCT(env_dct);
	::ZigZag(env_dct);

	// convert env_dct to integers
	int ofs = 0;

	for (y = 0; y < env_dct.sizey; y++)
	{
		for (int x = 0; x < env_dct.sizex; x++)
		{
			env_dct_int[0][ofs] = (int) env_dct.line[0][ofs].r;
			env_dct_int[1][ofs] = (int) env_dct.line[0][ofs].g;
			env_dct_int[2][ofs] = (int) env_dct.line[0][ofs].b;
			ofs++;
		}
	}
}
