#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>

#include <SDL/SDL.h>
#include <SDL/SDL_keysym.h>

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>

#include "floatimage.h"
#include "image.h"
#include "Render.h"
#include "glfont.h"
#include "xp_threads.h"

#define RADIANS_PER_DEGREE  0.01745329251994329576923690768
#define RADIANS(x)          RADIANS_PER_DEGREE * (x)

#define MAXPHI 32
#define MAXTHETA 64

#define PHISKIP 2
#define THETASKIP 2

static int WINDOW_WIDTH  = 1024;
static int WINDOW_HEIGHT =  768;
static int bpp           =   24;

static int  FULLSCREEN   =    0;
static int  NTHREADS     =    1;
static int  DELAY_FACTOR =   20;
#define     DETAIL_LEVELS     3
static int     FACES = 5;

static int done = 0;
static int face_invalid = 1;

//////////////////////////////////
// Faces

MipMapMonstaRenderer *monsta;

GLuint face_textures[DETAIL_LEVELS];
int    face_texture_size[DETAIL_LEVELS];

static int current_face = 0;

static int  MAX_FACES = 64;        // 64 faces ought to be enough

LDR_ReflectanceMapArray **face_data[64];
GLuint face_icon_names[64]; 

static float colorcorrect_r = 0.65;
static float colorcorrect_g = 0.68;
static float colorcorrect_b = 1.0;

//////////////////////////////////
// Environment maps

#define ENVIRONMENTS 6
static int   current_environment = 1;
static int   environment_spinning = 0;
static float environment_pos = 0.0f;
static float environment_lum = 0.5f;

floatimage *environments[ENVIRONMENTS];
GLuint env_names[ENVIRONMENTS];
GLuint center_env_name;

char env_text[ENVIRONMENTS][2][256];


//////////////////////////////////
// Lighting

static int lighting_rig_enable = 0;

typedef struct {
  float theta, phi;

  int   enabled;
  fRGB  color;
  float intensity;
  float spread;

  float last_color_x, last_color_y;

} light_t;

#define LIGHTS 3
#define AMBIENT (LIGHTS)
static light_t lights[LIGHTS + 1];

static int current_light = 0;


//////////////////////////////////
// Instructions screen

static int about_pressed = 0;
static int about_displayed = 0;

GLuint about_name;

//////////////////////////////////
// OpenGL stuff

GLFONT font;
GLuint sphere_dlist;
GLuint font_name, picker_name, credits_name;

// Did somebody order a unit sphere?
GLUquadricObj *glu_sphere;

// color picker
char *picker_data;


//////////////////////////////////
// UI handler stuff

static int mouse_x = 0;
static int mouse_y = 0;
static int mouse_dx = 0;
static int mouse_dy = 0;

SDL_Cursor *standard_cursor, *open_cursor, *closed_cursor;

static int dragging_environment = 0;
static int dragging_env_lum = 0;
static int dragging_intensity = 0;
static int dragging_diffuse = 0;
static int dragging_light = 0;

void draw_face_poly(int face, int level)
{
  float poly_width = 0.45f;

  float tex_x = (float)face_data[face][level]->sizex / (float)face_texture_size[level];
  float tex_y = (float)face_data[face][level]->sizey / (float)face_texture_size[level];

  float ratio  = (float)face_data[face][level]->sizey / 
    (float)face_data[face][level]->sizex;

  ratio *= poly_width;

  glPushMatrix();

  glTranslatef(0.0, 0.15, 0.0);

  glBegin( GL_QUADS );

  glTexCoord2f(0.01,0.01);
  glVertex2f(-poly_width, -ratio);

  glTexCoord2f(0.01,tex_y * 0.99);
  glVertex2f(-poly_width,ratio);

  glTexCoord2f(tex_x * 0.99,tex_y * 0.99);
  glVertex2f(poly_width,ratio);

  glTexCoord2f(tex_x * 0.99,0.01);
  glVertex2f(poly_width,-ratio);

  glEnd();

  glPopMatrix();
}

void draw_about() {

  // show about button
  glColor4ub(160, 160, 160, 255);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glVertex3f(-0.25, 0.97, 0.0);
  glVertex3f(-0.25, 0.85, 0.0);
  glVertex3f(0.3, 0.85, 0.0);
  glVertex3f(0.3, 0.97, 0.0);
  
  float inc = 0.008;
  if(about_pressed)
    inc = 0.0;

  glColor4ub(100, 100, 100, 255);
  glVertex3f(-0.25 - inc, 0.97 + inc, 0.0);
  glVertex3f(-0.25 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.3 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.3 - inc, 0.97 + inc, 0.0);
  glEnd();

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glFontBegin(&font);

  glColor4ub(255, 255, 255, 255);
  glFontSetScale(0.033, 0.03);
  glFontTextOut("Back", -0.08, 0.96, 0.05);

  glFontEnd();

  glDisable(GL_BLEND);
  glBindTexture(GL_TEXTURE_2D, about_name);

  float height = 1.73 * 0.5;
  float width = (4.0 / 3.0) * height;

  glBegin(GL_QUADS);

  glTexCoord2f(0.0,0.01);
  glVertex2f(-width, -height - 0.06);

  glTexCoord2f(0.0,0.99);
  glVertex2f(-width, height - 0.06);

  glTexCoord2f(1.0, 0.99);
  glVertex2f(width, height - 0.06);

  glTexCoord2f(1.0,0.01);
  glVertex2f(width, -height - 0.06);

  glEnd();
  
}


extern void draw_light(float r, float g, float b); // from mdl_light.cpp

void draw_lighting() {
  
  glPushMatrix();

  // establish a new coordinate system
  // just for the lighting rig
  glTranslatef(0.87, 0.45, 0.0);
  glScalef(0.12, 0.12, 0.12);

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);

  if(lights[0].enabled)
    glEnable(GL_LIGHT0);
  if(lights[1].enabled)
    glEnable(GL_LIGHT1);
  if(lights[2].enabled)
    glEnable(GL_LIGHT2);

  float pos[4];
  float col[4];

  pos[0] = 2.0 * cos(RADIANS(lights[0].phi)) * cos(RADIANS(lights[0].theta - 90.0));
  pos[1] = 2.0 * sin(RADIANS(lights[0].phi));
  pos[2] = 2.0 * cos(RADIANS(lights[0].phi)) * sin(RADIANS(lights[0].theta - 90.0));
  glLightfv(GL_LIGHT0, GL_POSITION, pos);

  // don't let the OGL display burn out too much
  col[0] = lights[0].color.r * lights[0].intensity * 0.25; 
  col[1] = lights[0].color.g * lights[0].intensity * 0.25;
  col[2] = lights[0].color.b * lights[0].intensity * 0.25;
  glLightfv(GL_LIGHT0, GL_DIFFUSE, col);

  pos[0] = 2.0 * cos(RADIANS(lights[1].phi)) * cos(RADIANS(lights[1].theta - 90.0));
  pos[1] = 2.0 * sin(RADIANS(lights[1].phi));
  pos[2] = 2.0 * cos(RADIANS(lights[1].phi)) * sin(RADIANS(lights[1].theta - 90.0));
  glLightfv(GL_LIGHT1, GL_POSITION, pos);

  // don't let the OGL display burn out too much
  col[0] = lights[1].color.r * lights[1].intensity * 0.25; 
  col[1] = lights[1].color.g * lights[1].intensity * 0.25;
  col[2] = lights[1].color.b * lights[1].intensity * 0.25;
  glLightfv(GL_LIGHT1, GL_DIFFUSE, col);

  pos[0] = 2.0 * cos(RADIANS(lights[2].phi)) * cos(RADIANS(lights[2].theta - 90.0));
  pos[1] = 2.0 * sin(RADIANS(lights[2].phi));
  pos[2] = 2.0 * cos(RADIANS(lights[2].phi)) * sin(RADIANS(lights[2].theta - 90.0));
  glLightfv(GL_LIGHT2, GL_POSITION, pos);

  // don't let the OGL display burn out too much
  col[0] = lights[2].color.r * lights[2].intensity * 0.25; 
  col[1] = lights[2].color.g * lights[2].intensity * 0.25;
  col[2] = lights[2].color.b * lights[2].intensity * 0.25;
  glLightfv(GL_LIGHT2, GL_DIFFUSE, col);

  if(lights[AMBIENT].enabled) {
     col[0] = lights[AMBIENT].color.r * lights[AMBIENT].intensity * 3.0;
     col[1] = lights[AMBIENT].color.g * lights[AMBIENT].intensity * 3.0;
     col[2] = lights[AMBIENT].color.b * lights[AMBIENT].intensity * 3.0;
  } else {
    col[0] = col[1] = col[2] = 0.0f;
  }
  
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, col);

  // draw the sphere
  glColor4ub(200,200,200,255);
  col[0] = 1.0; col[1] = 1.0; col[2] = 1.0;
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, col);

  glCallList(sphere_dlist);
  
  // reset state to render light icons

  glDisable(GL_LIGHTING);
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHT1);
  glDisable(GL_LIGHT2);
  glDisable(GL_LIGHT3);

  for(int i = 0; i < LIGHTS; i++) {
    glPushMatrix();
    glRotatef(lights[i].theta, 0.0, -1.0, 0.0);
    glRotatef(lights[i].phi, 1.0, 0.0, 0.0);

    if(i == current_light) {
      
      if(lights[i].enabled) {
	col[0] = 1.0; col[1] = 1.0; col[2] = 0.5; }
      else {
	col[0] = 0.5; col[1] = 0.4; col[2] = 0.1; }
      
      if(dragging_light) {
	col[0] = 1.0; col[1] = 1.0; col[2] = 1.0;
      }
      
    }
    else {
      if(lights[i].enabled) {
	col[0] = 0.25; col[1] = 0.15; col[2] = 1.0; }
      else {
	col[0] = 0.3; col[1] = 0.3; col[2] = 0.5; }
    }

    draw_light(col[0], col[1], col[2]);

    glPopMatrix();
  }

  
  // be kind, rewind
  glDisable(GL_DEPTH_TEST);
  glPopMatrix();

  // WIDGETRY

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);

  glFontBegin(&font);
  glFontSetScale(0.033, 0.03);
  glColor4ub(255,255,255,255);
  glFontTextOut("Diffusion:", 0.6, 0.05, 0.05);
  glFontTextOut("Intensity:", 0.6, -0.175, 0.05);
  glFontTextOut("Color:", 0.6, -0.37, 0.05);
  
  glFontSetScale(0.02, 0.015);
  glFontTextOut("SOFT", 0.515, -0.08, 0.05);
  glFontTextOut("HARD", 1.165, -0.08, 0.05);
  
  if(current_light == AMBIENT) {
    if(lights[AMBIENT].enabled)
      glColor4ub(255, 255, 200, 255);
    else
      glColor4ub(128, 128, 100, 255);
  }
  else {
    if(lights[AMBIENT].enabled)
      glColor4ub(100, 100, 255, 255);
    else
      glColor4ub(40, 40, 255, 255);
  }
  
  glFontTextOut("AMBIENT", 1.0, 0.12, 0.0);
  glFontEnd();

    // diffusion bar
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    
    // --- backing
    glColor4ub(50, 50, 50, 255);
    glVertex3f(0.672, -0.047, 0.0);
    glVertex3f(0.672, -0.153, 0.0);
    glVertex3f(1.138, -0.108, 0.0);
    glVertex3f(1.138, -0.092, 0.0);
    
    // --- gradient
    glColor4ub(0,0,0,255);
    glVertex3f(0.68, -0.055, 0.0);
    glVertex3f(0.68, -0.145, 0.0);
    glColor4ub(255,255,255,255);
    glVertex3f(1.13, -0.1, 0.0);
    glVertex3f(1.13, -0.1, 0.0);
    glEnd();

    // -- indicator
    if(current_light != AMBIENT) {

      glColor4ub(255, 255, 100, 255);
      // scale spread [2.0, 50.0] into slider range
      float marker = ((48.0 - (lights[current_light].spread-2.0)) * 0.45 * (1.0 / 48.0)) + 0.68;
      glBegin(GL_QUADS);
      glVertex3f(marker - 0.004, -0.055, 0.0);
      glVertex3f(marker - 0.004, -0.145, 0.0);
      glVertex3f(marker + 0.004, -0.145, 0.0);
      glVertex3f(marker + 0.004, -0.055, 0.0);
      glEnd();
    }

    else {
      glEnable(GL_TEXTURE_2D);
      glFontBegin(&font);
      glColor4ub(255, 255, 100, 255);
      glFontTextOut("AMBIENT", 0.7, -0.08, 0.05);
      glFontEnd();
    }
    
    // intensity bar
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    
    // --- backing
    glColor4ub(50, 50, 50, 255);
    glVertex3f(0.592, -0.267, 0.0);
    glVertex3f(0.592, -0.353, 0.0);
    glVertex3f(1.258, -0.353, 0.0);
    glVertex3f(1.256, -0.267, 0.0);
    
    // --- gradient
    glColor4ub(0,0,0,255);
    glVertex3f(0.6, -0.275, 0.0);
    glVertex3f(0.6, -0.345, 0.0);
    glColor4ub(255,255,255,255);
    glVertex3f(1.25, -0.345, 0.0);
    glVertex3f(1.25, -0.275, 0.0);
    glEnd();
    
    // --- indicator
    if(lights[current_light].enabled) {
      glColor4ub(255, 255, 100, 255);
      float marker = (lights[current_light].intensity * 0.65) + 0.6;
      glBegin(GL_QUADS);
      glVertex3f(marker - 0.004, -0.265, 0.0);
      glVertex3f(marker - 0.004, -0.355, 0.0);
      glVertex3f(marker + 0.004, -0.355, 0.0);
      glVertex3f(marker + 0.004, -0.265, 0.0);
      glEnd();
    }
    
    else {
      glEnable(GL_TEXTURE_2D);
      glFontBegin(&font);
      glColor4ub(255, 255, 100, 255);
      glFontTextOut("DISABLED", 0.8, -0.29, 0.05);
      glFontEnd();
    }
  
    
    // color picker
    glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, picker_name);
    glColor4ub(255, 255, 255, 255);
    
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0);
    glVertex3f(0.6, -0.475, 0.0);
    glTexCoord2f(0.0, 1.0);
    glVertex3f(0.6, -0.895, 0.0);
    glTexCoord2f(1.0, 1.0);
    glVertex3f(1.25, -0.895, 0.0);
    glTexCoord2f(1.0, 0.0);
    glVertex3f(1.25, -0.475, 0.0);
    glEnd();
    
    // color swatch

    glDisable(GL_TEXTURE_2D);

    glColor4f(lights[current_light].color.r,
	      lights[current_light].color.g,
	      lights[current_light].color.b,
	      1.0);
    
    glBegin(GL_QUADS);
    glVertex3f(0.83, -.375, 0.0);
    glVertex3f(0.83, -.455, 0.0);
    glVertex3f(0.95, -.455, 0.0);
    glVertex3f(0.95, -.375, 0.0);
    glEnd();
}




void draw_envselect()
{
  if(environment_spinning == 2) {
    environment_pos -= 0.5;
    face_invalid = 1;
  }
  else if(environment_spinning == 1) {
    environment_pos += 0.5;
    face_invalid = 1;
  }

  
  // text...
  glFontBegin(&font);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glColor4ub(255,255,255,255);
  glFontSetScale(0.03, 0.027);
  glFontTextOut("Env. Luminance", 0.7, -0.78, 0.05);
  glFontEnd();

  // environment luminance widget
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  
  // --- backing
  glColor4ub(50, 50, 50, 255);
  glVertex3f(0.65, -0.86, 0.0);
  glVertex3f(0.65, -0.94, 0.0);
  glVertex3f(1.25, -0.94, 0.0);
  glVertex3f(1.25, -0.86, 0.0);
    
  // --- gradient
  glColor4ub(0,0,0,255);
  glVertex3f(0.65 + 0.008, -0.86- 0.008, 0.0);
  glVertex3f(0.65 + 0.008, -0.94 + 0.008, 0.0);
  glColor4ub(255,255,255,255);
  glVertex3f(1.25 - 0.008, -0.94 + 0.008, 0.0);
  glVertex3f(1.25 - 0.008, -0.86 - 0.008, 0.0);
  glEnd();
    
  // --- indicator
  glColor4ub(255, 255, 100, 255);
  float marker = environment_lum * 0.6 + 0.65;

  glBegin(GL_QUADS);
  glVertex3f(marker - 0.004, -0.86, 0.0);
  glVertex3f(marker - 0.004, -0.94, 0.0);
  glVertex3f(marker + 0.004, -0.94, 0.0);
  glVertex3f(marker + 0.004, -0.86, 0.0);
  glEnd();

  // environment choice widget

  static float icon_height = 0.3;

  for(int i = 0; i < ENVIRONMENTS; i++) {
 
    float left    =  0.6;
    float right   =  0.9;
    float top     =  0.75 - (i/2) * (icon_height + 0.2);
    float bottom  =  top - icon_height;
    
    if(i % 2) {
      left  = 0.94;
      right = 1.24;
    }

    // draw border, lit if i == current environment

    if( i == current_environment )
      glColor4ub(250, 250, 250, 255);
    else
      glColor4ub(150, 150, 150, 255);
    
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glVertex3f(left - 0.008, top + 0.008, 0.0);
    glVertex3f(left - 0.008, bottom - 0.008, 0.0);
    glVertex3f(right + 0.008, bottom - 0.008, 0.0);
    glVertex3f(right + 0.008, top + 0.008, 0.0);
    glEnd();

    // draw icon

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, env_names[i]);

    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 1.0);
    glVertex3f(left, top, 0.0);
    glTexCoord2f(0.0, 0.0);
    glVertex3f(left, bottom, 0.0);
    glTexCoord2f(1.0, 0.0);
    glVertex3f(right, bottom, 0.0);
    glTexCoord2f(1.0, 1.0);
    glVertex3f(right, top, 0.0);
    glEnd();

    // draw text
    glEnable(GL_BLEND);
    glFontBegin(&font);
    glFontSetScale(0.016, 0.023);
    glFontTextOut(env_text[i][0], left, bottom - 0.03, 0.0);
    glFontTextOut(env_text[i][1], left, bottom - 0.09, 0.0);
    glFontEnd();
    glDisable(GL_BLEND);
  }
}




void draw_common()
{
  // show about button
  glColor4ub(160, 160, 160, 255);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glVertex3f(-0.25, 0.97, 0.0);
  glVertex3f(-0.25, 0.85, 0.0);
  glVertex3f(0.3, 0.85, 0.0);
  glVertex3f(0.3, 0.97, 0.0);
  
  float inc = 0.008;
  if(about_pressed)
    inc = 0.0;

  glColor4ub(100, 100, 100, 255);
  glVertex3f(-0.25 - inc, 0.97 + inc, 0.0);
  glVertex3f(-0.25 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.3 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.3 - inc, 0.97 + inc, 0.0);
  glEnd();


  // enable/disable lighting buttons

  glColor4ub(160, 160, 160, 255);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glVertex3f(0.60, 0.97, 0.0);
  glVertex3f(0.60, 0.85, 0.0);
  glVertex3f(0.90, 0.85, 0.0);
  glVertex3f(0.90, 0.97, 0.0);
  
  inc = 0.008;
  if( !lighting_rig_enable )
    inc = 0.0;

  glColor4ub(100, 100, 100, 255);
  glVertex3f(0.60 - inc, 0.97 + inc, 0.0);
  glVertex3f(0.60 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.90 - inc, 0.85 + inc, 0.0);
  glVertex3f(0.90 - inc, 0.97 + inc, 0.0);
  glEnd();

  //

  glColor4ub(160, 160, 160, 255);
  glBegin(GL_QUADS);
  glVertex3f(0.94, 0.97, 0.0);
  glVertex3f(0.94, 0.85, 0.0);
  glVertex3f(1.24, 0.85, 0.0);
  glVertex3f(1.24, 0.97, 0.0);
  
  inc = 0.008;
  if( lighting_rig_enable )
    inc = 0.0;

  glColor4ub(100, 100, 100, 255);
  glVertex3f(0.94 - inc, 0.97 + inc, 0.0);
  glVertex3f(0.94 - inc, 0.85 + inc, 0.0);
  glVertex3f(1.24 - inc, 0.85 + inc, 0.0);
  glVertex3f(1.24 - inc, 0.97 + inc, 0.0);
  glEnd();


  // credits
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, credits_name);
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0);
  glVertex3f(-1.325, 1.0, 0.0);
  glTexCoord2f(0.0, 1.0);
  glVertex3f(-1.325, 0.0, 0.0);
  glTexCoord2f(1.0, 1.0);
  glVertex3f(-0.325, 0.0, 0.0);
  glTexCoord2f(1.0, 0.0);
  glVertex3f(-0.325, 1.0, 0.0);
  glEnd();

  // text labels
  glFontBegin(&font);
  glFontSetScale(0.033, 0.03);
  glColor4ub(255, 255, 255, 255);
  glFontTextOut("Model:", -1.1, 0.3, 0.05);
  glFontSetScale(0.033, 0.03);
  glFontTextOut("About", -0.1, 0.96, 0.05);

  glFontSetScale(0.033 * 0.6, 0.03 * 0.8);

  if(lighting_rig_enable)
    glColor4ub(255, 255, 255, 255);
  else
    glColor4ub(255, 200, 0, 255);

  inc = 0.008;
  if( lighting_rig_enable )
    inc = 0.0;

  glFontTextOut("Environments", 0.610 + inc, 0.96 - inc, 0.0);

  if(lighting_rig_enable)
    glColor4ub(255, 200, 0, 255);
  else
    glColor4ub(255, 255, 255, 255);

  inc = 0.008;
  if( !lighting_rig_enable )
    inc = 0.0;

  glFontTextOut("Lights", 1.02 + inc, 0.96 - inc, 0.0);

  glFontEnd();

  glDisable(GL_BLEND);
  
  // center environment icon

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, center_env_name);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);


  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.98);
  glVertex3f(-0.5, -0.535, 0.0);
  glTexCoord2f(0.0, 0.02);
  glVertex3f(-0.5, -0.845, 0.0);
  glTexCoord2f(-1.0, 0.02);
  glVertex3f(0.5, -0.845, 0.0);
  glTexCoord2f(-1.0, 0.98);
  glVertex3f(0.5, -0.535, 0.0);
  glEnd();

  // arrows
  glDisable(GL_TEXTURE_2D);

  if(!lighting_rig_enable) {

    if(environment_spinning == 1)
      glColor4ub(255, 255, 200, 255);
    else
      glColor4ub(255, 255, 100, 255);

    glBegin(GL_TRIANGLES);
    glVertex3f(-0.508, -0.535, 0.0);
    glVertex3f(-0.558, -0.690, 0.0);
    glVertex3f(-0.508, -0.845, 0.0);
    
    if(environment_spinning == 2)
      glColor4ub(255, 255, 200, 255);
    else
      glColor4ub(255, 255, 100, 255);

    glVertex3f(0.508, -0.535, 0.0);
    glVertex3f(0.558, -0.690, 0.0);
    glVertex3f(0.508, -0.845, 0.0);
    glEnd();

  }


  // face choice widget

  static float icon_width = 0.2;

  for(int i = 0; i < FACES; i++) {
 
    float left    =  -1.3 + i * (icon_width + 0.06);
    float right   =  left + icon_width;
    float top     =  0.12;
    float bottom  = -0.22;

    if(i > 2) {
      left    =  -1.3 + (i-3) * (icon_width + 0.06);
      right   =  left + icon_width;

      if(FACES == 5) {
	left += 0.127; right += 0.127; }
      else if(FACES == 4) {
	left += 0.26; right += 0.26; }
      

      top = -0.27;
      bottom = -0.62;
    }

    // draw border, lit if i == current face

    if( i == current_face )
      glColor4ub(250, 250, 250, 255);
    else
      glColor4ub(150, 150, 150, 255);
    
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glVertex3f(left - 0.008, top + 0.008, 0.0);
    glVertex3f(left - 0.008, bottom - 0.008, 0.0);
    glVertex3f(right + 0.008, bottom - 0.008, 0.0);
    glVertex3f(right + 0.008, top + 0.008, 0.0);
    glEnd();

    // draw icon

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, face_icon_names[i]);

    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.9);
    glVertex3f(left, top, 0.0);
    glTexCoord2f(0.0, 0.1);
    glVertex3f(left, bottom, 0.0);
    glTexCoord2f(1.0, 0.1);
    glVertex3f(right, bottom, 0.0);
    glTexCoord2f(1.0, 0.9);
    glVertex3f(right, top, 0.0);
    glEnd();
  }
}


static floatimage translated_environment;

void redraw()
{
  static int last_known_good = 0;
  static int last_known_good_face = 0;
  static int first_run = 1;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(about_displayed) {
    draw_about();
    return;
  }

  draw_common();

  if(lighting_rig_enable) 
    draw_lighting();
  else
    draw_envselect();

  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  glDisable(GL_BLEND);

  // setup face texture
  if(first_run) {
    translated_environment.Allocate(MAXTHETA, MAXPHI);
    translated_environment.ShiftXFrom(*(environments[current_environment]), environment_pos);
    translated_environment.HalfSize();
    first_run = 0;
  }

  if(monsta->CalculationFinished()) {
    
    last_known_good = monsta->displevel;

    for(int i = 0; i < FACES; i++) {
      if(monsta->datalevel[ last_known_good ] == face_data[i][ last_known_good ])
	last_known_good_face = i;
    }

    glBindTexture(GL_TEXTURE_2D, face_textures[ last_known_good ] );
    
    glTexSubImage2D(GL_TEXTURE_2D, 0,
		    0, 0,
		    monsta->datalevel[ last_known_good ]->sizex,
		    monsta->datalevel[ last_known_good ]->sizey,
		    GL_RGB,
		    GL_UNSIGNED_BYTE,
		    monsta->imglevel[ last_known_good ].line[0] );

   if(face_invalid) 
     {
          face_invalid = 0;
	
	  
	  // prepare environment

	  // use only env. map, no lights
	  if( ! lighting_rig_enable) {
	    translated_environment.ShiftXFrom(*(environments[current_environment]), environment_pos * 2.0);
	    translated_environment *= (environment_lum * 3.0f);
	  }

	  // use only lights, no env. map
	  else {

	    translated_environment.Allocate(MAXTHETA, MAXPHI);
	    translated_environment.Clear(0.0, 0.0, 0.0);
	    for(int i = 0; i < LIGHTS; i++)
	      if(lights[i].enabled)
		translated_environment.AddLightPolar(lights[i].theta, lights[i].phi,
						     lights[i].color.r * lights[i].intensity * 14000.0,
						     lights[i].color.g * lights[i].intensity * 14000.0,
						     lights[i].color.b * lights[i].intensity * 14000.0,
						     lights[i].spread * 3.0);

	    if(lights[AMBIENT].enabled) {
	      fRGB add;
	      add.Set(lights[AMBIENT].color.r * lights[AMBIENT].intensity * 1200.0,
		      lights[AMBIENT].color.g * lights[AMBIENT].intensity * 1200.0,
		      lights[AMBIENT].color.b * lights[AMBIENT].intensity * 1200.0);

	      translated_environment += add;
	    }

	  }

	  // downsample and clamp for rendering
	

	  translated_environment.HalfSize();

	  // update center env texture
	  image ldr;
	  ldr.SetHDR((float*) translated_environment.line[0], 
		     translated_environment.sizex,
		     translated_environment.sizey,
		     0.001, 2.2);

	  glBindTexture(GL_TEXTURE_2D, center_env_name);
	  glTexSubImage2D(GL_TEXTURE_2D, 0,
			  0, 0,
			  MAXTHETA / THETASKIP, MAXPHI / PHISKIP,
			  GL_RGB, GL_UNSIGNED_BYTE,
			  ldr.line[0]);

	  translated_environment.ScaleChannels(colorcorrect_r,
					       colorcorrect_g,
					       colorcorrect_b);

	  // pixel values MUST be < 8192.0 for DCT code
	  translated_environment.ClampUpper(8192.0);


	  // get monsta going
	  monsta->SetMipMapLevel(0, face_data[current_face][0]);
	  monsta->SetMipMapLevel(1, face_data[current_face][1]);
	  monsta->SetMipMapLevel(2, face_data[current_face][2]);
	  monsta->NewCalculation(translated_environment);

	 
     }

   else // face not invalid, continue refining
     {
     monsta->Refine(translated_environment);
     }

  } // end if(calculationFinished)
  
  glBindTexture(GL_TEXTURE_2D, face_textures[ last_known_good ] );

  draw_face_poly(last_known_good_face, last_known_good);
}



















void init()
{
  // cursor setup
  standard_cursor = SDL_GetCursor();

  SDL_Surface *icon = SDL_LoadBMP("icon16.bmp");
  SDL_WM_SetIcon(icon, NULL);

  // cursor bitmaps, open + closed grabber
  static unsigned char open_mask[] = {
    0x00, 0x00,
    0x00, 0x00,
    0x03, 0xb8,
    0x1b, 0xf8,
    0x1f, 0xf8,
    0x1f, 0xf8,
    0x7f, 0xf8,
    0x7f, 0xff,
    0x7f, 0xff,
    0x7f, 0xff,
    0x7f, 0xff,
    0x3f, 0xfe,
    0x3f, 0xfc,
    0x1f, 0xf8,
    0x1f, 0xf0,
    0x1f, 0xf0
  };

  static unsigned char open_data[] = {
    0x00, 0x00,
    0x00, 0x00,
    0x03, 0xb8,
    0x1a, 0xe8,
    0x16, 0x48,
    0x12, 0x48,
    0x72, 0x48,
    0x5a, 0x4f,
    0x48, 0x19,
    0x40, 0x11,
    0x60, 0x03,
    0x20, 0x06,
    0x30, 0x0c,
    0x10, 0x18,
    0x10, 0x10,
    0x10, 0x10
  };

  static unsigned char closed_mask[] = {
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x01, 0xc0,
    0x03, 0xc0,
    0x1f, 0xf8,
    0x1f, 0xf8,
    0x3f, 0xfe,
    0x3f, 0xfe,
    0x3f, 0xfe,
    0x3f, 0xfe,
    0x3f, 0xfe,
    0x1f, 0xfc,
    0x1f, 0xf0,
    0x1f, 0xf0
  };

  static unsigned char closed_data[] = {
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x01, 0xc0,
    0x03, 0x40,
    0x1e, 0x78,
    0x12, 0x48,
    0x32, 0x4e,
    0x20, 0x1a,
    0x20, 0x02,
    0x20, 0x02,
    0x30, 0x06,
    0x10, 0x1c,
    0x10, 0x10,
    0x10, 0x10
  };

  open_cursor = SDL_CreateCursor(open_data, open_mask,
				 16, 16, 8, 8);

  closed_cursor = SDL_CreateCursor(closed_data, closed_mask,
				   16, 16, 8, 8);

  // set up the lighting sphere icon
  GLUquadricObj *glu_sphere = gluNewQuadric();
  gluQuadricDrawStyle(glu_sphere, (GLenum) GLU_FILL);
  gluQuadricNormals(glu_sphere, (GLenum) GLU_SMOOTH);
  sphere_dlist = glGenLists(1);
  glNewList(sphere_dlist, GL_COMPILE);
  gluSphere(glu_sphere, 1.0, 20, 20);
  glEndList();
  gluDeleteQuadric(glu_sphere);
 

  // initialize lights
  lights[0].theta = 95.0;
  lights[0].phi   = 45.0;
  lights[0].enabled = 1;
  lights[0].color.r = 0.5;
  lights[0].color.g = 0.5;
  lights[0].color.b = 0.5;
  lights[0].intensity = 0.5;
  lights[0].spread = 5.0;
  lights[0].last_color_x = 0.7;
  lights[0].last_color_y = -0.5;

  lights[1].theta = 67.0;
  lights[1].phi   = 67.0;
  lights[1].enabled = 1;
  lights[1].color.r = 0.2;
  lights[1].color.g = 0.3;
  lights[1].color.b = 0.5;
  lights[1].intensity = 0.2;
  lights[1].spread = 5.0;
  lights[1].last_color_x = 0.7;
  lights[1].last_color_y = -0.5;

  lights[2].theta = 300.0;
  lights[2].phi   = -20.0;
  lights[2].enabled = 1;
  lights[2].color.r = 0.5;
  lights[2].color.g = 0.3;
  lights[2].color.b = 0.2;
  lights[2].intensity = 0.2;
  lights[2].spread = 5.0;
  lights[2].last_color_x = 0.7;
  lights[2].last_color_y = -0.5;

  lights[AMBIENT].theta = 108.0;
  lights[AMBIENT].phi = -60.0;
  lights[AMBIENT].enabled = 1;
  lights[AMBIENT].color.r = 0.5;
  lights[AMBIENT].color.r = 0.5;
  lights[AMBIENT].color.g = 0.5;
  lights[AMBIENT].color.b = 0.5;
  lights[AMBIENT].intensity = 0.0;
  lights[AMBIENT].last_color_x = 0.7;
  lights[AMBIENT].last_color_y = -0.5;


  // general OpenGL init
  glClearColor(0.0,0.0,0.0,0.0);
  
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  
  glOrtho( -1.33, 1.333, -1.0, 1.0, -20.0, 20.0 );
  
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
  
  glDisable(GL_DEPTH_TEST);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

  // set up font
  glGenTextures(1, &font_name);
  glBindTexture(GL_TEXTURE_2D, font_name);
  glFontCreate(&font, "font.glf", font_name);

  // set up color picker graphic
  picker_data = (char*) malloc(256*256*3);

  FILE *f = fopen("color_picker.raw", "rb");
  if(!f) {
    fprintf(stderr, "Cannot open color picker image (color_picker.raw)\n");
    exit(-1);
  }

  fread(picker_data, 3, 256*256, f);
  fclose(f);

  glGenTextures(1, &picker_name);
  glBindTexture(GL_TEXTURE_2D, picker_name);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
		    256, 256,
		    GL_RGB, GL_UNSIGNED_BYTE,
		    picker_data);
  
  // DO NOT FREE picker_data!!! We use it later for figuring out
  // what color corresponds to what spot on the texture.

  // set up credit graphic
  image credit_image;
  credit_image.LoadBMP("credits.bmp");

  if(!credit_image.line) {
    fprintf(stderr, "Cannot open credits image (credits.bmp)\n");
    exit(-1);
  }

  credit_image.FlipVertical();

  glGenTextures(1, &credits_name);
  glBindTexture(GL_TEXTURE_2D, credits_name);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 1,
		    credit_image.sizex, credit_image.sizey,
		    GL_RGB, GL_UNSIGNED_BYTE,
		    credit_image.line[0]);


  // set up about graphic
  image about_image;
  about_image.LoadBMP("about.bmp");

  if(!about_image.line) {
    fprintf(stderr, "Cannot open \"about screen\" image (about.bmp)\n");
    exit(-1);
  }

  glGenTextures(1, &about_name);
  glBindTexture(GL_TEXTURE_2D, about_name);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 1,
		    about_image.sizex, about_image.sizey,
		    GL_RGB, GL_UNSIGNED_BYTE,
		    about_image.line[0]);



  ///////////////////////////////////////////////////////////
  // Set up enivronments (both HDR for calculation, and LDR for the GUI)

  glGenTextures(ENVIRONMENTS, env_names);

  if( chdir("environments") != 0) {
    fprintf(stderr, "Can't change to \"environments\" directory\n");
    SDL_Quit();
    exit(-1);
  }

  for(int i = 0; i < ENVIRONMENTS; i++) {
    environments[i] = new floatimage;
    char filename[256];

    sprintf(filename, "env%d.hdr", i);
    environments[i]->LoadHDR(filename);
    
    if(!environments[i]->line[0]) {
      printf("Can't load environment %s\n", filename);
      SDL_Quit();
      exit(-1);
    }
    
    environments[i]->ScaleMeanIntensityTo(19.0f);

    if(i > 2) // HACK XXX
      environments[i]->ScaleChannels(2.0, 2.0, 2.0);

    // construct 8-bit texture icon for the environment
    image foo;
    
    sprintf(filename, "env%d.bmp", i);
    foo.LoadBMP(filename);

    if(!foo.line[0]) {
      printf("Can't load environment icon %s\n", filename);
      SDL_Quit();
      exit(-1);
    }

    glBindTexture(GL_TEXTURE_2D, env_names[i]);
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
		      256, 256,
		      GL_RGB, GL_UNSIGNED_BYTE,
		      foo.line[0]);

    // load in text description
    sprintf(filename, "env%d.txt", i);
    FILE *f = fopen(filename, "rb");

    if(!f) {
      printf("Can't load environment description %s\n", filename);
      SDL_Quit();
      exit(-1);
    }

    // read two-line description
    fgets(env_text[i][0], 256, f);
    fgets(env_text[i][1], 256, f);
    fclose(f);
  }

  // set up center env icon

  char *bar = (char*) malloc(MAXTHETA / THETASKIP * MAXPHI / PHISKIP * 3);

  glGenTextures(1, &center_env_name);
  glBindTexture(GL_TEXTURE_2D, center_env_name);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
		    MAXTHETA / THETASKIP, MAXPHI / PHISKIP,
		    GL_RGB, GL_UNSIGNED_BYTE,
		    bar);

  free(bar);

  chdir("..");

  ///////////////////////////////////////////////////////////
  // Set up faces
  
  FACES = 0;

  if( chdir("faces") != 0 ) {
    fprintf(stderr, "Can't change to \"faces\" directory; does it exist?\n");
    SDL_Quit();
    exit(-1);
  }

  DIR *faces_dir = opendir(".");
  if(!faces_dir) {
    fprintf(stderr, "Error opening \"faces\" directory\n");
    SDL_Quit();
    exit(-1);
  }

  while(FACES < MAX_FACES) {

    struct dirent *dirent = readdir(faces_dir);
    if(!dirent) {
      if(FACES == 0) {
	fprintf(stderr, "No faces found (each face is a subdirectory of the \"faces\" directory)\n");
	SDL_Quit();
	exit(-1);
      }

      else
	break;
    }

    if(dirent->d_name[0] == (char) '.') // ignore ., .., etc
      continue;

    fprintf(stderr, "chdring to %s\n\n", dirent->d_name);

    if(chdir(dirent->d_name) != 0) {
      fprintf(stderr, "Can't change directory to faces/%s\n", dirent->d_name);
      SDL_Quit();
      exit(-1);
    }

    face_data[FACES] = (LDR_ReflectanceMapArray**) malloc(DETAIL_LEVELS * 
							  sizeof(LDR_ReflectanceMapArray*));

    for(int j = 0; j < DETAIL_LEVELS; j++) {
      char name[256];

      face_data[FACES][j] = new LDR_ReflectanceMapArray;
      
      sprintf(name, "level%d.map", j);

      face_data[FACES][j]->LoadMonsta(name);
      
      fprintf(stderr, "Loaded %s, size is %d x %d\n", name,
	      face_data[FACES][j]->sizex, face_data[FACES][j]->sizey);
      
      if(!face_data[FACES][j]->RLEbuf) {
	fprintf(stderr, "error loading reflectance data %s\n", name);

	char foo[256];
	getcwd(foo, 256);
	fprintf(stderr, "%s", foo);
	SDL_Quit();
	exit(-1);
      }
    }
    
    image im;
    im.LoadBMP("icon.bmp");
    
    if(!im.line) {
      fprintf(stderr, "Can't load face icon (icon.bmp)\n");
      SDL_Quit();
      exit(-1);
    }

    int tex_width = 1;
    while(tex_width < im.sizex)
      tex_width *= 2;
    int tex_height = 1;
    while(tex_height < im.sizey)
      tex_height *= 2;
    
    glGenTextures(1, &(face_icon_names[FACES])); 
    glBindTexture(GL_TEXTURE_2D, face_icon_names[FACES]);

    gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
		      tex_width, tex_height,
		      GL_RGB, GL_UNSIGNED_BYTE,
		      im.line[0]);

    
    chdir("..");
    FACES++;
  }

  closedir(faces_dir);

  
  monsta->SetMipMapLevel(0, face_data[0][0]);
  monsta->SetMipMapLevel(1, face_data[0][1]);
  monsta->SetMipMapLevel(2, face_data[0][2]);

  ///////////////////////////////////////////////////////////

  // find largest face dataset
  int face_width = 0;
  int face_height = 0;

  for(int i = 0; i < FACES; i++) {
    if(face_data[i][DETAIL_LEVELS - 1]->sizex > face_width)
      face_width = face_data[i][DETAIL_LEVELS - 1]->sizex;
    if(face_data[i][DETAIL_LEVELS - 1]->sizey > face_height)
      face_height = face_data[i][DETAIL_LEVELS - 1]->sizey;
  }

  fprintf(stderr, "Highest res of largest face dataset is %d x %d\n", face_width, face_height);

  printf("Launching %d render threads...\n", NTHREADS);
  monsta->InitThreads(NTHREADS);

  // set up face textures
  // one OpenGL texture per detail level
  // sized appropriately for the LARGEST face dataset
  // smaller datasets must be cropped via texture coordinates in draw_face_poly()

  int tex_width = 1;
  int tex_height = 1;

  while(tex_width < face_width)
    tex_width *= 2;
  while(tex_height < face_height)
    tex_height *= 2;

  int tex_size = (tex_width > tex_height? tex_width : tex_height);

  for(int i = 0; i < DETAIL_LEVELS; i++) {

    int size = tex_size >> (DETAIL_LEVELS - i - 1);

    face_texture_size[i] = size;
    
    fprintf(stderr, "LEVEL %d: TEXTURE SIZE %d x %d\n", i, size, size);
    
    char *tex_data = (char*) malloc(size * size * 3);
    
    glGenTextures(1, &(face_textures[i])); 
    glBindTexture(GL_TEXTURE_2D, face_textures[i]);

    gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
		      size, size,
		      GL_RGB, GL_UNSIGNED_BYTE,
		      tex_data);
    
    free(tex_data);
  }
 
}


// returns 1 if (mx,my) is in the rectangle
// returns 0 otherwise
int inrect(int mx, int my, float top, float bottom, float left, float right)
{
  float fmx = (float) mx / (float) WINDOW_WIDTH;
  float fmy = (float) my / (float) WINDOW_HEIGHT;

  // convert to OGL coordinate system
  // (left = -1.333, right = 1.333, btm = -1.0, top = 1.0)

  fmy = (-2.0 * fmy) + 1.0;
  fmx = (2.66666 * fmx) - 1.33333;
 

  if(fmx > left)
    if(fmx < right)
      if(fmy > bottom)
	if(fmy < top)
	  return 1;

  return 0;
}



void handle(SDL_Event *e)
{
  int double_click = 0;
  static int last_click = 0;
  int this_click;

  if(last_click == 0)
    last_click = SDL_GetTicks();


  switch(e->type) {

  case SDL_MOUSEBUTTONDOWN:

    this_click = SDL_GetTicks();

    if(this_click - last_click < 400) 
      if(((SDL_MouseButtonEvent*)e)->button == 1)
	double_click = 1;

    // check for selection of new environment

    for(int i = 0; i < ENVIRONMENTS; i++) {

      static float icon_height = 0.3;

      float left    =  0.6;
      float right   =  0.9;
      float top     =  0.75 - (i/2) * (icon_height + 0.2);
      float bottom  =  top - icon_height;
      
      if(i % 2) {
	left  = 0.94;
	right = 1.24;
      }
      

      if(inrect(mouse_x,
		mouse_y,
		top - 0.008, bottom + 0.008,
		left - 0.008, right + 0.008))
	{
	  current_environment = i;
	  environment_pos = 0.0;
	  environment_lum = 0.5;
	  face_invalid = 1;
	}
    }

    // check for selection of new face model

    for(int i = 0; i < FACES; i++) {

      static float icon_width = 0.2;

      float left    =  -1.3 + i * (icon_width + 0.06);
      float right   =  left + icon_width;
      float top     =  0.12;
      float bottom  = -0.22;
      
      if(i > 2) {
	left    =  -1.3 + (i-3) * (icon_width + 0.06);
	right   =  left + icon_width;

	
	if(FACES == 5) {
	  left += 0.127; right += 0.127; }
	else if(FACES == 4) {
	  left += 0.26; right += 0.26; }
	
	top = -0.27;
	bottom = -0.62;
      }

      if(inrect(mouse_x,
		mouse_y,
		top - 0.008, bottom + 0.008,
		left - 0.008, right + 0.008))
	{
	  current_face = i;
	  face_invalid = 1;
	}
    }

    // check environment spin buttons
    if(inrect(mouse_x, mouse_y, -0.535, -0.845, -0.558, -0.508)) {
      environment_spinning = 1;
      face_invalid = 1;
    }

    else if(inrect(mouse_x, mouse_y, -0.535, -0.845, 0.508, 0.558)) {
      environment_spinning = 2;
      face_invalid = 1;
    }
    
    // check environment/lighting toggle
    if(inrect(mouse_x, mouse_y, 0.97, 0.85, 0.6, 0.9) && lighting_rig_enable) {
      lighting_rig_enable = 0;
      face_invalid = 1;
    }

    else if(inrect(mouse_x, mouse_y, 0.97, 0.85, 0.94, 1.24) && !lighting_rig_enable) {
      lighting_rig_enable = 1;
      face_invalid = 1;
    }

    // check about button
    if(inrect(mouse_x,
	      mouse_y,
	      0.97, 0.85, -0.25, 0.3)
       && !dragging_light 
       && !dragging_environment 
       && !dragging_env_lum
       && !dragging_diffuse 
       && !dragging_intensity)
      {
	about_pressed = 1;
      }



    // check for diffusion change
    if(inrect(mouse_x,
	      mouse_y,
	      -0.065, -0.135, 0.68, 1.13) && lighting_rig_enable && (current_light != AMBIENT))
      {
	float xpos = (float) mouse_x / WINDOW_WIDTH;
	xpos = (2.66666 * xpos) - 1.33333;

	xpos -= 0.68; xpos /= (0.45 * (1.0 / 48.0));
	xpos = 48.0 - xpos; xpos += 1.0;

	lights[current_light].spread = xpos;

	face_invalid = 1;
      }

    // check for intensity change
    if(inrect(mouse_x,
	      mouse_y,
	      -0.275, -0.345, 0.6, 1.25) && lighting_rig_enable)
      {
	float xpos = (float) mouse_x / WINDOW_WIDTH;
	xpos = (2.66666 * xpos) - 1.33333;

	lights[current_light].intensity = (xpos - 0.6) / 0.65;

	face_invalid = 1;
      }

    // check for env lum
    if(inrect(mouse_x,
	      mouse_y,
	      -0.86, -0.94, 0.65, 1.25))
      {
	float xpos = (float) mouse_x / WINDOW_WIDTH;
	xpos = (2.66666 * xpos) - 1.33333;

	environment_lum = (xpos - 0.65) / 0.6;

	face_invalid = 1;
      }


    // check for color change
    if(inrect(mouse_x,
	      mouse_y,
	      -0.475, -0.895, 0.6, 1.25) && lighting_rig_enable)
      {
	float xpos = (float) mouse_x / WINDOW_WIDTH;
	xpos = (2.66666 * xpos) - 1.33333;
	
	lights[current_light].last_color_x = xpos;

	xpos -= 0.6; xpos /= 0.65; xpos *= 256.0;

	float ypos = (float) mouse_y / WINDOW_HEIGHT;
	ypos = (-2.0 * ypos) + 1.0;

	lights[current_light].last_color_y = ypos;

	ypos = -ypos - 0.475; ypos /= 0.42; ypos *= 256.0;

	unsigned int index = ((int) ypos % 256) * 256 * 3 + ((int)xpos % 256) * 3;

	unsigned char r = picker_data[index];
	unsigned char g = picker_data[index + 1];
	unsigned char b = picker_data[index + 2];
	
	lights[current_light].color.r = (float) r / 255.0f;
	lights[current_light].color.g = (float) g / 255.0f;
	lights[current_light].color.b = (float) b / 255.0f;

	
	face_invalid = 1;
      }
    
    if(lighting_rig_enable) {

      int found = 0;

    // check for light selection
    for(int i = 0; i < LIGHTS; i++) {

      float center_x = 0.87f;
      float center_y = 0.45f;

      center_x += 0.12 * 3.0 * sin(RADIANS(lights[i].theta)) * cos(RADIANS(lights[i].phi));
      center_y += 0.12 * 3.0 * sin(RADIANS(lights[i].phi));

      if(inrect(mouse_x,
		mouse_y,
		center_y + 0.08, center_y - 0.08,
		center_x - 0.08, center_x + 0.08)) {

	found = 1;
	current_light = i;
	dragging_light = 1;
	SDL_ShowCursor(0); // hide cursor

	if(double_click) {

	  // XXX
	  // disable light enabling/disabling for now;
	  // users seem to find it confusing

 	  // lights[i].enabled = ! lights[i].enabled;
	  // face_invalid = 1;
	}
	
      }

    }

    // also check ambient button
    if(inrect(mouse_x,
	      mouse_y,
	      0.2, 0.0, 1.0, 1.25) && !found) {
      current_light = AMBIENT;
      dragging_light = 1;

      if(double_click) {
	  lights[AMBIENT].enabled = ! lights[AMBIENT].enabled;
	  face_invalid = 1;
	}
    }

    }
    
    if(((SDL_MouseButtonEvent*)e)->button == 1 && !double_click)
      last_click = this_click;

    return;

  case SDL_KEYDOWN:

    if(((SDL_KeyboardEvent*)e)->keysym.sym == SDLK_ESCAPE)
      done = 1;

    return; 

  case SDL_QUIT:
    done = 1;
    return;
  }
}



void poll_mouse()
{
  int dx, dy;

  int buttons = SDL_GetMouseState(&mouse_x, &mouse_y);

  if(dragging_light) {

    if(mouse_x == 0)
      mouse_x = WINDOW_WIDTH - 2;
    else if(mouse_x >= (WINDOW_WIDTH -1)) {
      mouse_x = 1; }
    else if(mouse_y == 0)
      mouse_y = WINDOW_HEIGHT - 2;
    else if(mouse_y >= (WINDOW_HEIGHT -1))
      mouse_y = 1;
    else
      SDL_GetRelativeMouseState(&mouse_dx, &mouse_dy);

    SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
    SDL_WarpMouse(mouse_x, mouse_y);
    SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);
  }

  else
    SDL_GetRelativeMouseState(&mouse_dx, &mouse_dy);



  dx = mouse_dx; dy = mouse_dy;


  // environment hover
  if(!buttons) {
    int cursor_set = 0;

    if(inrect(mouse_x, mouse_y, -0.535, -0.845, -0.508, 0.508) &&
       !dragging_light && !dragging_intensity && !dragging_diffuse && !dragging_env_lum) {
      SDL_SetCursor(open_cursor);
      cursor_set = 1;
    }

    // light hover
    if(lighting_rig_enable) {

      for(int i = 0; i < LIGHTS; i++) {
	
	float center_x = 0.87f;
	float center_y = 0.45f;
	
	center_x += 0.12 * 3.0 * sin(RADIANS(lights[i].theta)) * cos(RADIANS(lights[i].phi));
	center_y += 0.12 * 3.0 * sin(RADIANS(lights[i].phi));
	
	if(inrect(mouse_x,
		  mouse_y,
		  center_y + 0.08, center_y - 0.08,
		  center_x - 0.08, center_x + 0.08)) {
	  
	  SDL_SetCursor(open_cursor);
	  cursor_set = 1;
	}
      }
    }

    if(!cursor_set)
      SDL_SetCursor(standard_cursor);

  }

  if(buttons) {

    
    // environment drag
    if(inrect(mouse_x, mouse_y, -0.535, -0.845, -0.508, 0.508)) {

      if(!dragging_light && !dragging_intensity && !dragging_diffuse && !dragging_env_lum) {
	environment_spinning = 0;
	dragging_environment = 1;
	SDL_SetCursor(closed_cursor);
      }
    }

    // intensity drag
    if(inrect(mouse_x,
	      mouse_y,
	      -0.225, -0.325, 0.6, 1.25) && lighting_rig_enable)
      {
	if(!dragging_light && !dragging_environment && !dragging_diffuse && !dragging_env_lum)
	  dragging_intensity = 1;
      }

    // env lum drag
    if(inrect(mouse_x,
	      mouse_y,
	      -0.86, -0.94, 0.65, 1.25))
      {
	if(!dragging_light && !dragging_environment && !dragging_diffuse && !dragging_intensity)
	  dragging_env_lum = 1;
      }  


    // diffuse drag
    if(inrect(mouse_x,
	      mouse_y,
	      -0.065, -0.135, 0.68, 1.13) && lighting_rig_enable && (current_light != AMBIENT))
      {
	if(!dragging_light && !dragging_environment && !dragging_intensity && !dragging_env_lum)
	  dragging_diffuse = 1;
      }


    // color picker
    if(inrect(mouse_x,
	      mouse_y,
	      -0.475, -0.895, 0.6, 1.25) && lighting_rig_enable)
      {

	if(!dragging_light && !dragging_environment && !dragging_intensity && !dragging_diffuse && !dragging_env_lum) {

	  float xpos = (float) (mouse_x) / WINDOW_WIDTH;
	  xpos = (2.66666 * xpos) - 1.33333;
	  
	  lights[current_light].last_color_x = xpos;
	  
	  xpos -= 0.6; xpos /= 0.65; xpos *= 256.0;
	  
	  float ypos = (float) (mouse_y) / WINDOW_HEIGHT;
	  ypos = (-2.0 * ypos) + 1.0;
	  
	  lights[current_light].last_color_y = ypos;
	  
	  ypos = -ypos - 0.475; ypos /= 0.42; ypos *= 256.0;
	  
	  unsigned int index = ((int) ypos % 256) * 256 * 3 + ((int)xpos % 256) * 3;
	  
	  unsigned char r = picker_data[index];
	  unsigned char g = picker_data[index + 1];
	  unsigned char b = picker_data[index + 2];
	  
	  lights[current_light].color.r = (float) r / 255.0f;
	  lights[current_light].color.g = (float) g / 255.0f;
	  lights[current_light].color.b = (float) b / 255.0f;
	
	  if(mouse_dx || mouse_dy)
	    face_invalid = 1;
	}

      }

  } 


  else {
    // mouse button not down

    if(dragging_environment) {

      if(inrect(mouse_x, mouse_y, -0.535, -0.845, -0.508, 0.508))
	SDL_SetCursor(open_cursor);
      else
	SDL_SetCursor(standard_cursor);

      SDL_ShowCursor(1);
      dragging_environment = 0;
    }


    dragging_env_lum = 0;
    dragging_intensity = 0;
    dragging_diffuse = 0;

    if(about_pressed) {
      about_pressed = 0;

      if(inrect(mouse_x, mouse_y,
		0.97, 0.85, -0.25, 0.3))
	{
	  about_displayed = ! about_displayed;
	  face_invalid = 1;
	}
    }

    if(dragging_light) {

      dragging_light = 0;

      // warp cursor to light location
      float center_x = 0.87f;
      float center_y = 0.45f;

      center_x += 0.12 * 3.0 * sin(RADIANS(lights[current_light].theta)) * cos(RADIANS(lights[current_light].phi));
      center_y += 0.12 * 3.0 * sin(RADIANS(lights[current_light].phi));

      if(current_light == AMBIENT) {
	center_x = 1.1;
	center_y = 0.12;
      }

      // transform to window coordinates
      center_x += 1.333333; center_x /= 2.66666; center_x *= WINDOW_WIDTH;
      center_y -= 1.0; center_y /= -2.0; center_y *= WINDOW_HEIGHT;

      mouse_x = (int) center_x; mouse_y = (int) center_y;

      SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
      SDL_WarpMouse(mouse_x, mouse_y);
      SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);

      SDL_SetCursor(standard_cursor);
      SDL_ShowCursor(1);
    }
  }


  if(!mouse_dx && !mouse_dy)
    return;


  if(dragging_environment) {
    environment_pos += (-mouse_dx) / 8.0f;

    if(lighting_rig_enable) {
      for(int i = 0; i < LIGHTS; i++) {
	lights[i].theta += (-mouse_dx) / 2.0f;
      }
    }

    face_invalid = 1;
    return;
  }
 
  if(dragging_diffuse) {

    float xpos = (float) mouse_x / WINDOW_WIDTH;
    xpos = (2.66666 * xpos) - 1.33333;


    xpos -= 0.68; xpos /= (0.45 * (1.0 / 48.0));
    xpos = 48.0 - xpos; xpos += 1.0;
    
    if(xpos < 1.0) xpos = 1.0;
    else if(xpos > 50.0) xpos = 50.0;

    lights[current_light].spread = xpos;
    
    face_invalid = 1;

  }


  if(dragging_intensity) {
    float xpos = (float) mouse_x / WINDOW_WIDTH;

    xpos = (2.66666 * xpos) - 1.33333;

    lights[current_light].intensity = (xpos - 0.6) / 0.65;

    if(lights[current_light].intensity < 0.0)
      lights[current_light].intensity = 0.0;
    else if(lights[current_light].intensity > 1.0)
      lights[current_light].intensity = 1.0;

    face_invalid = 1;
    return;
  }

  if(dragging_env_lum) {
    float xpos = (float) mouse_x / WINDOW_WIDTH;

    xpos = (2.66666 * xpos) - 1.33333;

    environment_lum = (xpos - 0.65) / 0.6;

    if(environment_lum < 0.0)
      environment_lum = 0.0;
    else if(environment_lum > 1.0)
      environment_lum = 1.0;

    face_invalid = 1;
    return;
  }


  if(dragging_light) {

    // LMB + RMB changes diffusion
    if((buttons & SDL_BUTTON(1)) && (buttons & SDL_BUTTON(3)) || (buttons & SDL_BUTTON(2))) 
      {
	if(dx > -100 && dx < 100)
	  lights[current_light].spread -= ((float) dx) / 3.5;

	if(lights[current_light].spread > 50.0)
	  lights[current_light].spread = 50.0;
	if(lights[current_light].spread < 2.0)
	  lights[current_light].spread = 2.0;
      }

    // RMB changes intensity
    else if(buttons & SDL_BUTTON(3))
      {
	if(dx > -100 && dx < 100)
	  lights[current_light].intensity += ((float) dx) / 100.0;

	if(lights[current_light].intensity > 1.0)
	  lights[current_light].intensity = 1.0;
	if(lights[current_light].intensity < 0.0)
	  lights[current_light].intensity = 0.0;
      }

    // LMB changes position
    else if(buttons & SDL_BUTTON(1) && current_light != AMBIENT)
      {
	if(dx > -200 && dx < 200)
	  lights[current_light].theta += ((float) -dx) / 2.0f;
	
	if(lights[current_light].theta < 0.0f)
	  lights[current_light].theta += 360.0f;
	if(lights[current_light].theta > 360.0f)
	  lights[current_light].theta -= 360.0f;
	
	if(dy > -200 && dy < 200)
	  lights[current_light].phi   += ((float) -dy) / 2.0f;
	
	if(lights[current_light].phi > 75.0f)
	  lights[current_light].phi  = 75.0f;
	if(lights[current_light].phi < -30.0f)
	  lights[current_light].phi  = -30.0f;
      }

    face_invalid = 1;
    return;
  }

}

void configure(int argc, char *argv[])
{
  FILE *f = fopen("config.txt", "rb");

  if(!f) {
    fprintf(stderr, "Warning - can't open config.txt\n");
    goto skip_file;
  }

  while(!feof(f)) {
    char buf[256];
    char arg[256];
    char foo[256];

    fgets(buf, 256, f);
    fgets(arg, 256, f);
    fgets(foo, 256, f); // suck up empty line

    if(strstr(buf, "threads"))
      NTHREADS = atoi(arg);
    else if(strstr(buf, "delay"))
      DELAY_FACTOR = atoi(arg);
    else if(strstr(buf, "width"))
      WINDOW_WIDTH = atoi(arg);
    else if(strstr(buf, "height"))
      WINDOW_HEIGHT = atoi(arg);
    else if(strstr(buf, "faces"))
      MAX_FACES = atoi(arg);
    else if(strstr(buf, "bpp"))
      bpp = atoi(arg);
    else if(strstr(buf, "r"))
      colorcorrect_r = atof(arg);
    else if(strstr(buf, "g"))
      colorcorrect_g = atof(arg);
    else if(strstr(buf, "b"))
      colorcorrect_b = atof(arg);
  }

  fclose(f);

 skip_file:

  for(int i = 1; i < argc; i++) {

    if(!strcmp(argv[i], "--threads")) {
      NTHREADS = atoi(argv[++i]);
      if(NTHREADS > 32)
	NTHREADS = 32;
      if(NTHREADS < 1)
	NTHREADS = 1;
    }

    else if(!strcmp(argv[i], "--delay")) {
      DELAY_FACTOR = atoi(argv[++i]);
      if(DELAY_FACTOR < 0)
	DELAY_FACTOR = 0;
    }

    else if(!strcmp(argv[i], "--width"))
      WINDOW_WIDTH = atoi(argv[++i]);

    else if(!strcmp(argv[i], "--height"))
      WINDOW_HEIGHT = atoi(argv[++i]);

    else if(!strcmp(argv[i], "--fullscreen"))
      FULLSCREEN = 0; // XXX never fullscreen (pointer warping is b0rken)
    
    else if(!strcmp(argv[i], "--r"))
      colorcorrect_r = atof(argv[++i]);

    else if(!strcmp(argv[i], "--g"))
      colorcorrect_g = atof(argv[++i]);

    else if(!strcmp(argv[i], "--b"))
      colorcorrect_b = atof(argv[++i]);

    else if(!strcmp(argv[i], "--bpp")) {
      int new_bpp = atoi(argv[++i]);
      if(( new_bpp == 24) || (new_bpp == 16) )
	bpp = new_bpp;
    }

    else if(!strcmp(argv[i], "--faces")) {
      MAX_FACES = atoi(argv[++i]);
    }

    else if(!strcmp(argv[i], "--help")) {
      printf("Usage: %s [--delay N] [--threads N]\n"
	     "\t --delay       : Length of time GUI thread sleeps each loop, in msec\n"
	     "\t --threads     : Number of rendering threads\n"
	     "\t --faces       : Number of faces (~40MB RAM required per face)\n"
	     "\t --width       : Width of window, in pixels\n"
	     "\t --height      : Height of window, in pixels\n"
	     "\t --bpp         : Color depth, 16 or 24 bits\n"
	     "\t --r, --g, --b : Global color correction\n"
	     "\t --fullscreen  : Run full-screen\n"    ,argv[0]);
      exit(1);
    }
 }
}

int main( int argc, char* argv[] )
{
  thread_priority_high();

  monsta = new MipMapMonstaRenderer;

  configure(argc, argv);

  if( SDL_Init(SDL_INIT_VIDEO) < 0 )
    exit(-1);

  Uint32 video_flags = SDL_OPENGL;

  if(FULLSCREEN)
    video_flags |= SDL_FULLSCREEN;

  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

  if ( SDL_SetVideoMode( WINDOW_WIDTH,  WINDOW_HEIGHT, bpp, video_flags ) == NULL ) {
    fprintf(stderr, "Couldn't set GL mode: %s\n", SDL_GetError());
    SDL_Quit();
    exit(1);
  }
	
  glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

  init();
  
  while(!done) {
    SDL_Event event;

    mouse_dx = mouse_dy = 0; // reset in case no mouse motion event is delivered

    while(SDL_PollEvent(&event))
      handle(&event);

    
    poll_mouse();

    redraw();
    SDL_GL_SwapBuffers();

#ifdef _WIN32
    Sleep(DELAY_FACTOR);
#else
    usleep(DELAY_FACTOR * 1000);
#endif

  }

  // delete monsta;      heh, oh boy... (~monsta() is segfaulting somewhere)

  SDL_Quit();
  return 1;
}
