/* Copyright (c) 1996 Regents of the University of California */

#ifndef lint
static char SCCSid[] = "@(#)header.c 2.8 7/27/96 LBL";
#endif

/*
 *  header.c - routines for reading and writing information headers.
 *
 *	8/19/88
 *
 *  newheader(t,fp)	start new information header identified by string t
 *  isheadid(s)		returns true if s is a header id line
 *  headidval(r,s)	copy header identifier value in s to r
 *  printargs(ac,av,fp) print an argument list to fp, followed by '\n'
 *  isformat(s)		returns true if s is of the form "FORMAT=*"
 *  formatval(r,s)	copy the format value in s to r
 *  fputformat(s,fp)	write "FORMAT=%s" to fp
 *  getheader(fp,f,p)	read header from fp, calling f(s,p) on each line
 *  globmatch(pat, str)	check for glob match of str against pat
 *
 *  To copy header from input to output, use getheader(fin, fputs, fout)
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define	 MAXLINE	512

#ifndef BSD
#define	 index	strchr
#endif

/* extern char  *index(); */

char  HDRSTR[] = "#?";		/* information header magic number */

char  FMTSTR[] = "FORMAT=";	/* format identifier */

void
newheader(char  *s, register FILE  *fp)		/* identifying line of information header */
{
	fputs(HDRSTR, fp);
	fputs(s, fp);
	putc('\n', fp);
}


int
headidval(register char  *r,register char  *s)			/* get header id (return true if is id) */
{
	register char  *cp = HDRSTR;

	while (*cp) if (*cp++ != *s++) return(0);
	if (r == NULL) return(1);
	while (*s) *r++ = *s++;
	*r = '\0';
	return(1);
}


int
isheadid(char  *s)			/* check to see if line is header id */
{
	return(headidval(NULL, s));
}

void
printargs(int ac, char  **av, register FILE  *fp)		/* print arguments to a file */
{
	int  quote;

	while (ac-- > 0) {
		if (index(*av, ' ') != NULL) {		/* quote it */
			if (index(*av, '\'') != NULL)
				quote = '"';
			else
				quote = '\'';
			putc(quote, fp);
			fputs(*av++, fp);
			putc(quote, fp);
		} else
			fputs(*av++, fp);
		putc(ac ? ' ' : '\n', fp);
	}
}


int
formatval(register char  *r, register char  *s)			/* get format value (return true if format) */
{
	register char  *cp = FMTSTR;

	while (*cp) if (*cp++ != *s++) return(0);
	while (isspace(*s)) s++;
	if (!*s) return(0);
	if (r == NULL) return(1);
	do
		*r++ = *s++;
	while(*s && !isspace(*s));
	*r = '\0';
	return(1);
}


int
isformat(char  *s)			/* is line a format line? */
{
	return(formatval(NULL, s));
}

void
fputformat(char  *s, FILE  *fp)		/* put out a format value */
{
	fputs(FMTSTR, fp);
	fputs(s, fp);
	putc('\n', fp);
}


int getheader(FILE  *fp)		/* get header from file */
{
	char  buf[MAXLINE];

	for ( ; ; ) {
		buf[MAXLINE-2] = '\n';
		if (fgets(buf, MAXLINE, fp) == NULL)
			return(-1);
		if (buf[0] == '\n')
			return(0);
#ifdef MSDOS
		if (buf[0] == '\r' && buf[1] == '\n')
			return(0);
#endif
		if (buf[MAXLINE-2] != '\n') {
			ungetc(buf[MAXLINE-2], fp);	/* prevent false end */
			buf[MAXLINE-2] = '\0';
		}
		/*		if (f != NULL)
			(*f)(buf, p); */
	}
}


struct check {
	FILE	*fp;
	char	fs[64];
};


static void
mycheck(char  *s, register struct check  *cp)			/* check a header line for format info. */
{
	if (!formatval(cp->fs, s) && cp->fp != NULL)
		fputs(s, cp->fp);
}


int
globmatch(char	*pat, char	*str)		/* check for glob match of str against pat */
{
	register char	*p = pat, *s = str;

	do {
		switch (*p) {
		case '?':			/* match any character */
			if (!*s++)
				return(0);
			break;
		case '*':			/* match any string */
			while (p[1] == '*') p++;
			do
				if ( (p[1]=='?' || p[1]==*s) &&
						globmatch(p+1,s) )
					return(1);
			while (*s++);
			return(0);
		case '\\':			/* literal next */
			p++;
		/* fall through */
		default:			/* normal character */
			if (*p != *s)
				return(0);
			s++;
			break;
		}
	} while (*p++);
	return(1);
}
